from drivnal.constants import *
from drivnal.client import Client
from drivnal.task import Task
from drivnal.event import Event
import drivnal.utils as utils
from drivnal import server
import os
import flask
import time

@server.app.route('/task/<volume_id>', methods=['GET'])
def task_get(volume_id):
    client = Client()
    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    tasks = []

    for task in reversed(volume.get_tasks()):
        task_data = {
            'id': task.id,
            'volume': volume.id,
            'volume_name': volume.name,
            'type': task.type,
            'state': task.state,
            'time': task.time,
            'has_log': False,
        }

        if task.log_path:
            if os.path.isfile(task.log_path):
                task_data['has_log'] = True

        if task.snapshot_id:
            task_data['snapshot_id'] = task.snapshot_id

        tasks.append(task_data)

    return utils.jsonify(tasks)

@server.app.route('/task/<volume_id>/<task_id>', methods=['PUT'])
def task_put(volume_id, task_id):
    client = Client()
    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    task = Task(id=task_id.encode())

    if 'abort' in flask.request.json and flask.request.json['abort']:
        task.abort()

    return utils.jsonify({})

@server.app.route('/task/<volume_id>/<task_id>', methods=['DELETE'])
def task_delete(volume_id, task_id):
    client = Client()
    task = Task(id=task_id.encode())
    task.remove()

    return utils.jsonify({})



from constants import *
from core_volume import CoreVolume
from local_snapshot import LocalSnapshot
import os
import logging
import subprocess

logger = logging.getLogger(APP_NAME)

class LocalVolume(CoreVolume):
    SnapshotClass = LocalSnapshot

    def _get_log_dir(self):
        return os.path.join(self.path, LOG_DIR)

    def _get_auto_excludes(self):
        source_path = self.source_path
        if source_path[-1] != os.sep:
            source_path += os.sep

        # If volume is a subdirectory of source path exclude path
        if os.path.commonprefix([source_path, self.path]) == source_path:
            logger.debug('Auto excluding volume path from snapshot. %r' % {
                'volume_id': self.id,
            })
            path = os.sep + os.path.normpath(self.path.replace(
                source_path, '', 1)) + os.sep
            return [path]

    def _get_mount(self, path):
        output = subprocess.check_output(['df', '-P', path])
        return output.strip().split(' ')[-1]

    def _move_available(self, source_path, destination_path):
        return self._get_mount(source_path) == self._get_mount(
            destination_path)

    def _copy_available(self, source_path, destination_path):
        return True

    def get_space_free(self):
        stat = os.statvfs(self.path)
        return float(stat.f_bavail) / stat.f_blocks

    def list_path(self, path, files=True, dirs=True):
        path = os.path.join(self.path, path)
        names = []

        if not os.path.isdir(path):
            return names

        for name in os.listdir(path):
            if files and dirs:
                pass
            elif not files and dirs:
                if not os.path.isdir(os.path.join(path, name)):
                    continue
            elif files and not dirs:
                if not os.path.isfile(os.path.join(path, name)):
                    continue
            else:
                continue
            names.append(name)

        return names



import flask
import json

def format_bytes(bytes):
    bytes = float(bytes)
    if bytes < 1024:
        return '%.2f bytes' % bytes
    elif bytes < 1048576:
        return '%.2f kB' % (bytes / 1024)
    elif bytes < 1073741824:
        return '%.2f MB' % (bytes / 1048576)
    elif bytes < 1099511627776:
        return '%.2f GB' % (bytes / 1073741824)
    else:
        return '%.2f TB' % (bytes / 1099511627776)

def format_seconds(seconds):
    seconds = float(seconds)
    if seconds < 60:
        return str(int(seconds)) + ' Seconds'
    elif seconds < 3600:
        return '%.1f Minutes' % (seconds / 60)
    elif seconds < 86400:
        return '%.1f Hours' % (seconds / 3600)
    else:
        return '%.1f Days' % (seconds / 86400)

def jsonify(data=None, status_code=None):
    if not isinstance(data, basestring):
        data = json.dumps(data)

    callback = flask.request.args.get('callback', False)
    if callback:
        data = '%s(%s)' % (callback, data)
        mimetype = 'application/javascript'
    else:
        mimetype = 'application/json'

    response = flask.Response(response=data, mimetype=mimetype)

    if status_code is not None:
        response.status_code = status_code

    return response



APP_NAME = 'drivnal'
APP_NAME_FORMATED = 'Drivnal'
CONF_FILENAME = '%s.conf' % APP_NAME

SAVED = 'saved'
UNSAVED = 'unsaved'

FAILED = 'failed'
REMOVING = 'removing'
ABORTED = 'aborted'
ABORTING = 'aborting'
PENDING = 'pending'
WARNING = 'warning'
COMPLETE = 'complete'

CREATE_SNAPSHOT = 'create_snapshot'
REMOVE_SNAPSHOT = 'remove_snapshot'
RESTORE_OBJECT = 'restore_object'
MOVE_VOLUME = 'move_volume'

VOLUMES_UPDATED = 'volumes_updated'
SNAPSHOTS_UPDATED = 'snapshots_updated'
TASKS_UPDATED = 'tasks_updated'
ORIGIN_UPDATED = 'origin_updated'

SNAPSHOT_DIR = 'snapshots'
LOG_DIR = 'logs'

EVENT_DB_TTL = 61000
TASK_DB_MAX = 100
MAX_TEXT_SIZE = 20971520

SMTP_TIMEOUT = 30
SMTP_FROM_ADDR = 'notify@drivnal.com'
SMTP_SUBJECT = 'Drivnal Notification'

MATCH_ALL = 'match_all'
MATCH_BOTH = 'match_both'
MATCH_PREFIX = 'match_prefix'
MATCH_SUFFIX = 'match_suffix'

DEFAULT_MAX_PRUNE = 0.8
DEFAULT_DB_PATH = '/var/lib/drivnal/drivnal.db'
DEFAULT_WWW_PATH = '/usr/share/drivnal/www'

DEFAULT_ROOT_EXCLUDES = [
    '/dev/*',
    '/proc/*',
    '/sys/*',
    '/tmp/*',
    '/run/*',
    '/mnt/*',
    '/media/*',
    '/lost+found',
]

DIR_MIME_TYPE = 'inode/directory'

RSYNC_WARN_EXIT_CODES = [
    23, # Partial transfer from error
    24, # Partial transfer from vanished files
]

MIN_UNITS = ['m', 'min', 'minute']
HOUR_UNITS = ['h', 'hr', 'hour']
DAY_UNITS = ['d', 'day']
WEEKDAY_UNITS = {
    'monday': 0,
    'mon': 0,
    'mo': 0,
    'm': 0,
    'tuesday': 1,
    'tues': 1,
    'tue': 1,
    'tu': 1,
    't': 1,
    'wednesday': 2,
    'wed': 2,
    'we': 2,
    'w': 2,
    'thursday': 3,
    'thur': 3,
    'thu': 3,
    'th': 3,
    'friday': 4,
    'fri': 4,
    'fr': 4,
    'f': 4,
    'saturday': 5,
    'sat': 5,
    'sa': 5,
    's': 5,
    'sunday': 6,
    'sun': 6,
    'su': 6,
}

EXTENSION_TYPES = {
    'abap': 'abap',
    'as': 'actionscript',
    'adb': 'ada',
    'ads': 'ada',
    'x86': 'assembly_x86',
    'ahk': 'autohotkey',
    'bat': 'batchfile',
    'cmd': 'batchfile',
    'c': 'c_cpp',
    'w': 'c_cpp',
    'cpp': 'c_cpp',
    'cxx': 'c_cpp',
    'h': 'c_cpp',
    'hh': 'c_cpp',
    'hxx': 'c_cpp',
    'h++': 'c_cpp',
    'tcc': 'c_cpp',
    'cs': 'csharp',
    'csx': 'csharp',
    'cob': 'cobol',
    'cbl': 'cobol',
    'ccp': 'cobol',
    'cobol': 'cobol',
    'cpy': 'cobol',
    'clj': 'clojure',
    'cljs': 'clojure',
    'cljx': 'clojure',
    'coffee': 'coffee',
    '_coffee': 'coffee',
    'cson': 'coffee',
    'iced': 'coffee',
    'cfm': 'coldfusion',
    'cfc': 'coldfusion',
    'css': 'css',
    'd': 'd',
    'di': 'd',
    'dart': 'dart',
    'diff': 'diff',
    'mustache': 'django',
    'jinja': 'django',
    'dot': 'dot',
    'gv': 'dot',
    'erl': 'erlang',
    'hrl': 'erlang',
    'fth': 'forth',
    '4th': 'forth',
    'glsl': 'glsl',
    'fp': 'glsl',
    'frag': 'glsl',
    'geom': 'glsl',
    'glslv': 'glsl',
    'shader': 'glsl',
    'vert': 'glsl',
    'go': 'golang',
    'groovy': 'groovy',
    'haml': 'haml',
    'haml.deface': 'haml',
    'html.haml.deface': 'haml',
    'hs': 'haskell',
    'hsc': 'haskell',
    'hx': 'haxe',
    'hxsl': 'haxe',
    'htm': 'html',
    'html': 'html',
    'xhtml': 'html',
    'ini': 'ini',
    'cfg': 'ini',
    'prefs': 'ini',
    'properties': 'ini',
    'java': 'java',
    'js': 'javascript',
    '_js': 'javascript',
    'bones': 'javascript',
    'jake': 'javascript',
    'jsfl': 'javascript',
    'jsm': 'javascript',
    'jss': 'javascript',
    'pac': 'javascript',
    'sjs': 'javascript',
    'ssjs': 'javascript',
    'json': 'json',
    'jq': 'jsoniq',
    'jqy': 'jsoniq',
    'jsp': 'jsp',
    'jsx': 'jsx',
    'jl': 'julia',
    'aux': 'latex',
    'dtx': 'latex',
    'ins': 'latex',
    'ltx': 'latex',
    'sty': 'latex',
    'toc': 'latex',
    'less': 'less',
    'lisp': 'lisp',
    'asd': 'lisp',
    'lsp': 'lisp',
    'ny': 'lisp',
    'podsl': 'lisp',
    'ls': 'livescript',
    '_ls': 'livescript',
    'lua': 'lua',
    'nse': 'lua',
    'rbxs': 'lua',
    'mk': 'makefile',
    'mak': 'makefile',
    'md': 'markdown',
    'markdown': 'markdown',
    'mkd': 'markdown',
    'mkdown': 'markdown',
    'ron': 'markdown',
    'matlab': 'matlab',
    'm': 'objectivec',
    'mm': 'objectivec',
    'ml': 'ocaml',
    'mli': 'ocaml',
    'mll': 'ocaml',
    'mly': 'ocaml',
    'eliomi': 'ocaml',
    'pas': 'pascal',
    'dfm': 'pascal',
    'lpr': 'pascal',
    'pl': 'perl',
    'nqp': 'perl',
    'perl': 'perl',
    'ph': 'perl',
    'plx': 'perl',
    'pm6': 'perl',
    'pod': 'perl',
    'psgi': 'perl',
    'php': 'php',
    'php3': 'php',
    'php4': 'php',
    'php5': 'php',
    'phpt': 'php',
    'aw': 'php',
    'ctp': 'php',
    'phtml': 'php',
    'txt': 'plain_text',
    'ps1': 'powershell',
    'pro': 'prolog',
    'py': 'python',
    'gyp': 'python',
    'pyt': 'python',
    'pyw': 'python',
    'wsgi': 'python',
    'xpy': 'python',
    'r': 'r',
    'rhtml': 'rhtml',
    'rb': 'ruby',
    'rbw': 'ruby',
    'rbx': 'ruby',
    'builder': 'ruby',
    'gemspec': 'ruby',
    'god': 'ruby',
    'irbrc': 'ruby',
    'podspec': 'ruby',
    'rbuild': 'ruby',
    'ru': 'ruby',
    'thor': 'ruby',
    'watchr': 'ruby',
    'rs': 'rust',
    'sass': 'sass',
    'scala': 'scala',
    'scm': 'scheme',
    'sls': 'scheme',
    'ss': 'scheme',
    'scss': 'scss',
    'sh': 'sh',
    'tmux': 'sh',
    'bashrc': 'sh',
    'sql': 'sql',
    'svg': 'svg',
    'tcl': 'tcl',
    'tex': 'tex',
    'textile': 'textile',
    'tmsnippet': 'tmsnippet',
    'tmcommand': 'tmsnippet',
    'tmlanguage': 'tmsnippet',
    'tmpreferences': 'tmsnippet',
    'tmtheme': 'tmsnippet',
    'toml': 'toml',
    'twig': 'twig',
    'ts': 'typescript',
    'vbs': 'vbscript',
    'v': 'verilog',
    'sv': 'verilog',
    'svh': 'verilog',
    'vh': 'verilog',
    'xml': 'xml',
    'axml': 'xml',
    'ccxml': 'xml',
    'dita': 'xml',
    'ditamap': 'xml',
    'ditaval': 'xml',
    'glade': 'xml',
    'grxml': 'xml',
    'kml': 'xml',
    'mxml': 'xml',
    'plist': 'xml',
    'pt': 'xml',
    'rdf': 'xml',
    'rss': 'xml',
    'scxml': 'xml',
    'tml': 'xml',
    'ui': 'xml',
    'vxml': 'xml',
    'wsdl': 'xml',
    'wxi': 'xml',
    'wxl': 'xml',
    'wxs': 'xml',
    'x3d': 'xml',
    'xaml': 'xml',
    'xlf': 'xml',
    'xliff': 'xml',
    'xmi': 'xml',
    'xsd': 'xml',
    'xul': 'xml',
    'zcml': 'xml',
    'xquery': 'xquery',
    'xq': 'xquery',
    'xqy': 'xquery',
    'yml': 'yaml',
    'reek': 'yaml',
    'yaml': 'yaml',
};

# Mime match string, syntax type
MIME_TYPES = [
    ('actionscript', 'actionscript'),
    ('typescript', 'typescript'),
    ('powershell', 'powershell'),
    ('coldfusion', 'coldfusion'),
    ('makefile', 'makefile'),
    ('plain', 'plain_text'),
    ('xhtml', 'xhtml'),
    ('shell', 'sh'),
    ('text', 'plain_text'),
    ('bash', 'sh'),
    ('html', 'html'),
    ('diff', 'diff'),
    ('xml', 'xml'),
];

# Filename match string, syntax type
FILENAME_TYPES = [
    ('vagrantfile', 'ruby', MATCH_ALL),
    ('pkgbuild', 'bash', MATCH_ALL),
    ('bash', 'bash', MATCH_ALL),
    ('xml', 'xml', MATCH_ALL),
    ('rc', 'bash', MATCH_SUFFIX),
];

VOLUME_NOT_FOUND = 'volume_not_found'
VOLUME_NOT_FOUND_MSG = 'Volume not found'

SNAPSHOT_NOT_FOUND = 'snapshot_not_found'
SNAPSHOT_NOT_FOUND_MSG = 'Snapshot not found'

TASK_NOT_FOUND = 'task_not_found'
TASK_NOT_FOUND_MSG = 'Task not found'

PATH_NOT_FOUND = 'path_not_found'

FILE_NOT_FOUND = 'file_not_found'
FILE_NOT_FOUND_MSG = 'File not found'



from constants import *
from exceptions import *
from task import Task
from event import Event
import os
import time
import shlex
import logging
import subprocess

logger = logging.getLogger(APP_NAME)

class ExecTask(Task):
    def _pre_aborted(self):
        pass

    def _post_aborted(self):
        pass

    def _abort_process(self):
        for i in xrange(10):
            self.process.terminate()
            logger.debug('Terminating task process. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_type': self.type,
                'pid': self.process.pid,
            })
            time.sleep(1)
            if self.process.poll() is not None:
                break

        if self.process.poll() is None:
            for i in xrange(30):
                self.process.kill()
                logger.debug('Killing task process. %r' % {
                    'volume_id': self.volume_id,
                    'snapshot_id': self.snapshot_id,
                    'task_type': self.type,
                    'pid': self.process.pid,
                })
                time.sleep(0.5)
                if self.process.poll() is not None:
                    break

        if self.process.poll() is None:
            logger.error('Failed to abort task process. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_type': self.type,
                'pid': self.process.pid,
            })

        logger.warning('Task aborted. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot_id,
            'task_type': self.type,
        })

        self._pre_aborted()
        self.aborted()
        self._post_aborted()

    def _exec(self, args, env=None, shell=False):
        self.process = subprocess.Popen(args, env=env, shell=shell)
        return_code = None

        while True:
            poll = self.process.poll()
            if poll is not None:
                return_code = poll
                break
            if self.state == ABORTING:
                self._abort_process()
                return
            time.sleep(0.5)

        return return_code



from constants import *
from core_snapshot import CoreSnapshot
import os
import logging

logger = logging.getLogger(APP_NAME)

class LocalSnapshot(CoreSnapshot):
    def _get_path(self):
        dir_name = str(self.id)
        if self.state != COMPLETE:
            dir_name = '%s.%s' % (dir_name, self.state)
        return os.path.join(self.volume.path, SNAPSHOT_DIR, dir_name) + os.sep

    def _get_log_path(self):
        return os.path.join(self.volume.log_dir, 'snapshot_%s.log' % self.id)

    def _setup_snapshot(self, last_snapshot):
        snapshots_path = os.path.dirname(self.path)
        if not os.path.isdir(snapshots_path):
            logger.debug('Creating volume snapshots directory. %r' % {
                'volume_id': self.volume.id,
                'snapshot_id': self.id,
            })
            os.mkdir(snapshots_path)

        logs_path = os.path.dirname(self.log_path)
        if not os.path.isdir(logs_path):
            logger.debug('Creating volume logs directory. %r' % {
                'volume_id': self.volume.id,
                'snapshot_id': self.id,
            })
            os.mkdir(logs_path)

        if last_snapshot:
            return ['--link-dest=%s' % last_snapshot.path]

    def set_state(self, state):
        if self.state == state:
            return

        orig_state = self.state
        orig_path = self.path
        self.state = state

        try:
            os.rename(orig_path, self.path)
        except OSError:
            logger.exception('Unable to change snapshot state. %r' % {
                'volume_id': self.volume.id,
                'snapshot_id': self.id,
                'state': self.state,
            })
            self.state = orig_state



from constants import *
import os
import logging

logger = logging.getLogger(APP_NAME)
_RESERVED_ATTRIBUTES = ['bool_options', 'int_options', 'float_options',
    'path_options', 'str_options', 'list_options', 'all_options']

class Config:
    bool_options = []
    int_options = []
    float_options = []
    path_options = []
    str_options = []
    list_options = []

    def __init__(self, path=None):
        self.all_options = self.bool_options + self.int_options + \
            self.float_options + self.path_options + self.str_options
        self._conf_path = path
        self._loaded = False
        self.set_state(SAVED)

    def __setattr__(self, name, value):
        if name in _RESERVED_ATTRIBUTES:
            pass
        elif name in self.all_options:
            self.set_state(UNSAVED)
        self.__dict__[name] = value

    def __getattr__(self, name):
        if name in _RESERVED_ATTRIBUTES:
            pass
        elif name in self.all_options:
            if not self._loaded:
                self.load()
            if name not in self.__dict__:
                return [] if name in self.list_options else None
        elif name not in self.__dict__:
            raise AttributeError('Config instance has no attribute %r' % name)
        return self.__dict__[name]

    def get_state(self):
        return self._state

    def set_state(self, state):
        self._state = state

    def get_path(self):
        return self._conf_path

    def set_path(self, path):
        self._conf_path = path

    def _encode_line(self, name, value):
        if name in self.bool_options:
            value = 'true' if value else 'false'
        elif name in self.list_options:
            value = ','.join(value)

        return '%s=%s\n' % (name, value)

    def _decode_bool(self, value):
        value = value.lower()
        if value in ['true', 't', 'yes', 'y']:
            return True
        elif value in ['false', 'f', 'no', 'n']:
            return False
        else:
            raise ValueError('Value is not boolean')

    def _decode_int(self, value):
        return int(value)

    def _decode_float(self, value):
        return float(value)

    def _decode_path(self, value):
        try:
            value = os.path.normpath(value)
        except AttributeError:
            logger.error('Failed to normalize path. %r' % {
                'path': value,
            })
        return value

    def _decode_list(self, value):
        return filter(None, value.split(','))

    def _decode_line(self, line):
        line_split = line.split('=')
        name = line_split[0]
        value = '='.join(line_split[1:])

        if value:
            if name not in self.all_options:
                raise ValueError('Unknown option')

            if not value:
                raise ValueError('Empty option')

            if name in self.list_options:
                values = self._decode_list(value)

                decoder = None
                if name in self.int_options:
                    decoder = self._decode_int
                elif name in self.float_options:
                    decoder = self._decode_float
                elif name in self.bool_options:
                    decoder = self._decode_bool
                elif name in self.path_options:
                    decoder = self._decode_path

                if decoder:
                    for i, value in enumerate(values):
                        values[i] = decoder(value)

                value = values
            elif name in self.str_options:
                pass
            elif name in self.int_options:
                value = self._decode_int(value)
            elif name in self.float_options:
                value = self._decode_float(value)
            elif name in self.bool_options:
                value = self._decode_bool(value)
            elif name in self.path_options:
                value = self._decode_path(value)

        else:
            value = [] if name in self.list_options else None

        return name, value

    def load(self, merge=False):
        logger.debug('Loading config.')
        self._loaded = True

        try:
            with open(self._conf_path) as config:
                for line in config:
                    line = line.rstrip('\n')

                    if line.strip() == '':
                        continue
                    elif line[0] == '#':
                        continue
                    elif '=' in line:
                        pass
                    else:
                        logger.warning('Ignoring invalid line. %r' % {
                            'line': line,
                        })
                        continue

                    try:
                        name, value = self._decode_line(line)
                        if merge and name in self.__dict__:
                            continue
                        self.__dict__[name] = value
                    except ValueError:
                        logger.warning('Ignoring invalid line. %r' % {
                            'line': line,
                        })
        except IOError:
            if not merge:
                raise

    def commit(self):
        logger.debug('Committing config.')
        if not self._loaded:
            self.load(True)

        with open(self._conf_path, 'w') as config:
            for name in self.all_options:
                if name not in self.__dict__:
                    continue
                value = self.__dict__[name]
                if value is None:
                    continue
                config.write(self._encode_line(name, value))

        self.set_state(SAVED)



from constants import *
import smtplib
import time
import threading
import logging

logger = logging.getLogger(APP_NAME)

class Messenger:
    def __init__(self, volume):
        self._thread = None
        self.volume_id = volume.id

        self.email = volume.email
        if volume.email_host:
            smtp_host_split = volume.email_host.split(':')
            if len(smtp_host_split) > 1:
                self.smtp_host = smtp_host_split[0]
                try:
                    self.smtp_port = int(smtp_host_split[1])
                except TypeError:
                    self.smtp_port = None
                    logger.error('Failed to convert smtp port number in ' + \
                        'config. %r' % {
                            'volume_id': self.volume_id,
                            'smtp_port': smtp_host_split[1],
                        })
            else:
                self.smtp_host = smtp_host_split[0]
                self.smtp_port = None
        self.smtp_user = volume.email_user
        self.smtp_pass = volume.email_pass
        self.smtp_ssl = volume.email_ssl

    def _send(self, message):
        if not self.email:
            return
        message = 'From %s\r\nTo: %s\r\nSubject:%s\r\n\r\n%s' % (
            SMTP_FROM_ADDR, self.email, SMTP_SUBJECT, message)

        host = self.smtp_host or None
        port = self.smtp_port or None
        username = self.smtp_user or None
        password = self.smtp_pass or None

        logger.debug('Sending smtp message. %r' % {
            'volume_id': self.volume_id,
            'smtp_host': host,
            'smtp_port': port,
            'smtp_ssl': self.smtp_ssl,
            'smtp_user': username,
            'smtp_pass': password,
            'message': message,
        })

        try:
            if self.smtp_ssl:
                server = smtplib.SMTP_SSL(host, port)
            else:
                server = smtplib.SMTP(host, port)

            if username or password:
                server.login(username, password)
            server.sendmail(SMTP_FROM_ADDR, [self.email], message)
            server.quit()
        except:
            logger.exception('Failed to send email. %r' % {
                'volume_id': self.volume_id,
                'smtp_host': host,
                'smtp_port': port,
                'smtp_ssl': self.smtp_ssl,
                'smtp_user': username,
                'smtp_pass': password,
                'message': message,
            })

    def send(self, message):
        if self._thread:
            self._thread.join()
        self._thread = threading.Thread(target=self._send, args=(message,))
        self._thread.start()



from constants import *
import os
import stat
import urllib
import subprocess
import logging

logger = logging.getLogger(APP_NAME)

class Object:
    def __init__(self, path):
        self.name = os.path.basename(path)
        self.path = path
        self.syntax = None
        self.readable = False

        if os.path.isdir(self.path):
            self.type = DIR_MIME_TYPE
            self.size = None
            self.time = None
        else:
            try:
                file_stat = os.stat(self.path)
                self.size = file_stat.st_size
                self.time = int(file_stat.st_mtime)
                self.readable = bool(file_stat.st_mode & stat.S_IROTH)
            except OSError:
                self.size = None
                self.time = None
            self.type = None

    def get_mime_type(self):
        if not self.type:
            try:
                # TODO Follow symlinks
                self.type = subprocess.check_output(['file',
                    '--mime-type', '--brief', self.path]).strip()
            except subprocess.CalledProcessError, error:
                logger.debug('File mime-type call failed. %r' % {
                    'return_code': error.returncode,
                    'output': error.output,
                })
        return self.type

    def get_syntax(self):
        if self.type == DIR_MIME_TYPE:
            return

        extension = os.path.splitext(self.path)[1].lower().replace('.', '')
        if extension in EXTENSION_TYPES:
            self.syntax = EXTENSION_TYPES[extension]
            return self.syntax

        for match, syntax, match_type in FILENAME_TYPES:
            name = self.name.lower()

            if match_type == MATCH_ALL:
                if match in name:
                    self.syntax = syntax
                    return self.syntax
                continue

            if match_type == MATCH_BOTH or match_type == MATCH_PREFIX:
                if name.startswith(match):
                    self.syntax = syntax
                    return self.syntax
                if match_type == MATCH_PREFIX:
                    continue

            if name.endswith(match):
                self.syntax = syntax
                return self.syntax

        if self.get_mime_type():
            for mime_type, syntax in MIME_TYPES:
                if mime_type in self.type.lower():
                    self.syntax = syntax
                    return self.syntax

    def read(self):
        # TODO
        # if self.readable:
        #     return

        if not self.syntax:
            return

        try:
            with open(self.path) as file_data:
                return file_data.read()
        except OSError, error:
            logger.warning('Failed to read file. %r' % {
                'error': error,
            })
            return

    @staticmethod
    def get_objects(path):
        objects = []
        object_paths = []

        if path:
            for name in os.listdir(path):
                object_path = os.path.join(path, name)
                object = Object(object_path)
                objects.append(object)
                if not object.type:
                    object_paths.append(object_path)

        # Attempt to get all mime types with one call get_syntax will
        # call again if this fails
        try:
            # TODO Follow symlinks
            mime_types = subprocess.check_output(['file',
                '--mime-type', '--brief'] + object_paths).splitlines()
        except subprocess.CalledProcessError, error:
            logger.debug('File mime-type call failed. %r' % {
                'return_code': error.returncode,
                'output': error.output,
            })

        try:
            for object in objects:
                if not object.type:
                    object.type = mime_types.pop(0)
        except IndexError:
            logger.error('File mime-type call index error.')

        for object in objects:
            object.get_syntax()

        return objects



from constants import *
from config import Config
from local_volume import LocalVolume
from event import Event
from drivnal import server
import os
import copy
import logging

logger = logging.getLogger(APP_NAME)

class Client:
    def get_volume(self, id):
        for volume in self.get_volumes():
            if volume.id == id:
                return volume

    def get_volumes(self):
        logger.debug('Loading client volumes.')
        volumes = []

        for i, path in enumerate(server.app_db.get('system', 'volumes')):
            try:
                path = os.path.normpath(path)
            except AttributeError:
                logger.error('Failed to normalize volume path. %r' % {
                    'path': path,
                })
                continue

            try:
                volume = LocalVolume(self, path)
                volumes.append(volume)
            except IOError:
                logger.debug('Failed to load volume. %r' % {
                    'path': path,
                })

        return volumes

    def add_volume(self, path):
        logger.debug('Adding volume.')

        if not os.path.isdir(path):
            os.mkdir(path)
        volume = LocalVolume(self, path, create=True)

        logger.debug('Adding volume path to database. %r' % {
            'volume_id': volume.id,
            'path': path,
        })
        server.app_db.set('system', 'volumes', volume.path, None)

        Event(type=VOLUMES_UPDATED)

        return volume

    def remove_volume(self, volume_id):
        logger.debug('Removing volume. %r' % {
            'volume_id': volume_id,
        })

        for volume in self.get_volumes():
            if volume.id == volume_id:
                server.app_db.remove('system', 'volumes', volume.path)
                Event(type=VOLUMES_UPDATED)
                break



"""Toggle scheduling for vagrant stress testing volumes."""
import fileinput

for i in xrange(1, 9):
    file_path = '/media/volumes/backup%s/drivnal.conf' % i
    for line in fileinput.input(file_path, inplace=True):
        if 'schedule=none' in line:
            line = 'schedule=1minute'
        elif 'schedule=' in line:
            line = 'schedule=none'
        print line.rstrip('\n')



from drivnal.constants import *
from drivnal.client import Client
from drivnal.task import Task
from drivnal.object import Object
import drivnal.utils as utils
from drivnal import server
import flask
import time
import os

@server.app.route('/restore/<volume_id>/<int:snapshot_id>', methods=['POST'])
def restore_post(volume_id, snapshot_id):
    client = Client()

    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    snapshot = volume.get_snapshot(snapshot_id)

    if not snapshot:
        return utils.jsonify({
            'error': SNAPSHOT_NOT_FOUND,
            'error_msg': SNAPSHOT_NOT_FOUND_MSG,
        }, 404)

    source_paths = flask.request.json['source_paths']
    destination_path = os.path.join(
        volume.source_path, flask.request.json['destination_path'])
    source_objects = []

    for path in source_paths:
        source_objects.append(Object(os.path.join(snapshot.path, path)))

    volume.restore_object(source_objects, destination_path)

    return utils.jsonify({})



from constants import *
from drivnal import server
from database_object import DatabaseObject
import logging
import time
import uuid

logger = logging.getLogger(APP_NAME)

class Event(DatabaseObject):
    db = server.mem_db
    column_family = 'events'
    str_columns = ['volume_id', 'type']
    int_columns = ['time']
    cached_columns = ['volume_id', 'type', 'time']
    required_columns = ['type', 'time']

    def __init__(self, id=None, volume_id=None, type=None):
        DatabaseObject.__init__(self)

        if id is None:
            self.id = uuid.uuid4().hex
            self.type = type
            self.time = int(time.time() * 1000)
            self.volume_id = None
        else:
            self.id = id

        if volume_id is not None:
            self.volume_id = volume_id

    @staticmethod
    def clean_database():
        cur_time = int(time.time() * 1000)

        events_query = Event.db.get(Event.column_family)
        for event_id in events_query:
            event = events_query[event_id]

            # Skip broken events
            if not DatabaseObject.validate(Event, event_id, event):
                continue

            event['time'] = int(event['time'])

            # Remove events after ttl
            if (cur_time - event['time']) > EVENT_DB_TTL:
                logger.debug('Removing event past ttl from database. %r' % {
                    'event_id': event_id,
                })
                Event.db.remove(Event.column_family, event_id)
                continue

    @staticmethod
    def get_events(last_time=0):
        events = []
        events_dict = {}
        events_time = []
        cur_time = int(time.time() * 1000)

        logger.debug('Getting global events. %r' % {
            'last_time': last_time,
        })

        events_query = Event.db.get(Event.column_family)
        for event_id in events_query:
            event = events_query[event_id]

            # Skip broken events
            if not DatabaseObject.validate(Event, event_id, event):
                continue

            event['time'] = int(event['time'])

            if event['time'] <= last_time:
                continue

            # Prevent events with the same time from breaking sorted list,
            # event that is sent to client will always have original time
            while True:
                if event['time'] not in events_time:
                    break
                event['time'] += 1

            events_dict[event['time']] = Event(id=event_id)
            events_time.append(event['time'])

        for event_time in sorted(events_time):
            events.append(events_dict[event_time])

        return events



from constants import *
from object import Object
import os
import logging

class Bucket:
    def __init__(self, path):
        self.path = path

    def get_object(self, path):
        return Object(os.path.join(self.path, path))

    def get_objects(self, path=None):
        if path:
            return Object.get_objects(os.path.join(self.path, path))
        else:
            return Object.get_objects(self.path)



from constants import *
from event import Event
from database_object import DatabaseObject
import threading
import logging
import time
import uuid
import os

logger = logging.getLogger(APP_NAME)
_task_threads = {}

class Task(DatabaseObject):
    column_family = 'tasks'
    str_columns = ['volume_id', 'type', 'state', 'log_path']
    int_columns = ['time', 'snapshot_id']
    cached_columns = ['volume_id', 'type', 'time', 'snapshot_id']
    required_columns = ['volume_id', 'type', 'state', 'time']

    def __init__(self, id=None, volume=None, snapshot=None):
        DatabaseObject.__init__(self)

        if id is None:
            self.id = uuid.uuid4().hex
            # Transfer the class attribute to the database
            self.type = self.type
            self.state = None
            self.time = None
            self.volume_id = None
            _task_threads[self.id] = None

        else:
            self.id = id

        if volume is None:
            self.volume = None
        else:
            self.volume_id = volume.id
            self.volume = volume

        if snapshot is not None:
            self.snapshot_id = snapshot.id
            self.snapshot = snapshot

        self.update()

    def __getattr__(self, name):
        if name == 'thread':
            if self.id not in _task_threads:
                return None
            return _task_threads[self.id]
        else:
            return DatabaseObject.__getattr__(self, name)

    def __setattr__(self, name, value):
        if name == 'thread':
            if value is None:
                if self.id in _task_threads:
                    del _task_threads[self.id]
            else:
                _task_threads[self.id] = value
        else:
            DatabaseObject.__setattr__(self, name, value)

        if name == 'state' and value:
            Event(volume_id=self.volume_id, type=TASKS_UPDATED)

    def log_read(self):
        if not self.log_path:
            return
        try:
            with open(self.log_path) as file_data:
                return file_data.read()
        except OSError, error:
            logger.warning('Failed to read task log file. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
                'error': error,
            })
            return

    def log_size(self):
        if not self.log_path:
            return
        try:
            return os.path.getsize(self.log_path)
        except OSError, error:
            logger.warning('Failed to get task log file size. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
                'error': error,
            })
            return

    def abort(self):
        logger.debug('Aborting task. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })
        self.state = ABORTING
        Event(volume_id=self.volume_id, type=TASKS_UPDATED)

    def aborted(self):
        logger.debug('Task aborted. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })
        self.state = ABORTED
        self.thread = None
        Event(volume_id=self.volume_id, type=TASKS_UPDATED)

    def remove(self):
        logger.debug('Removing task. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })
        # Volume id will be removed in next call
        volume_id = self.volume_id
        self.db.remove(self.column_family, self.id)
        Event(volume_id=volume_id, type=TASKS_UPDATED)

    def run(self, *args, **kwargs):
        raise NotImplementedError('Derived class must override run method')

    def _run(self, *args, **kwargs):
        try:
            self.run(*args, **kwargs)
            if self.state == PENDING:
                self.state = COMPLETE
            elif self.state == ABORTING:
                self.state = ABORTED
            self.thread = None
        except:
            logger.exception('Task failed. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
            })
            self.state = FAILED
            self.thread = None
        finally:
            if self.state == PENDING:
                self.state = FAILED
                self.thread = None

    def start(self, *args, **kwargs):
        self.volume_id = self.volume_id
        self.time = int(time.time())
        self.state = PENDING
        self.thread = threading.Thread(
            target=self._run, args=args, kwargs=kwargs)
        self.thread.start()

    def join(self):
        if not self.thread:
            return
        self.thread.join()

    def update(self):
        if self.state not in [PENDING, ABORTING]:
            return

        if not self.thread or not self.thread.is_alive():
            logger.warning('Task failed, thread ended unexpectable. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
            })
            self.state = FAILED

    @staticmethod
    def clean_database():
        tasks_dict = {}
        tasks_time = []

        tasks_query = Task.db.get(Task.column_family)
        for task_id in tasks_query:
            task = tasks_query[task_id]

            # Skip broken events
            if not DatabaseObject.validate(Task, task_id, task):
                continue

            task['time'] = int(task['time'])

            task_time_id = '%s-%s' % (task['time'], task_id)
            tasks_dict[task_time_id] = task_id
            tasks_time.append(task_time_id)

        prune_count = len(tasks_query) - TASK_DB_MAX
        if prune_count <= 0:
            return

        # Remove tasks after limit is reached
        tasks_time = sorted(tasks_time)
        for i in xrange(prune_count):
            task_id = tasks_dict[tasks_time.pop(0)]
            logger.debug('Max task count reached removing task ' + \
                    'from database. %r' % {
                'task_id': task_id,
            })
            Task.db.remove(Task.column_family, task_id)

    @staticmethod
    def _get_tasks(volume, type, states=[]):
        tasks = []
        tasks_dict = {}
        tasks_time = []

        logger.debug('Getting tasks for volume. %r' % {
            'volume_id': volume.id,
        })

        tasks_query = Task.db.get(Task.column_family)
        for task_id in tasks_query:
            task = tasks_query[task_id]

            # Skip broken events
            if not DatabaseObject.validate(Task, task_id, task):
                continue

            task['time'] = int(task['time'])
            if type and task['type'] != type:
                continue
            if task['volume_id'] != volume.id:
                continue
            if states and task['state'] not in states:
                continue
            if not states and not task['state']:
                continue

            # Task update occurs here recheck state
            task = Task(id=task_id)
            if states and task.state not in states:
                continue
            task_time_id = '%s-%s' % (task.time, task_id)
            tasks_dict[task_time_id] = task
            tasks_time.append(task_time_id)

        for task_time_id in sorted(tasks_time):
            tasks.append(tasks_dict[task_time_id])

        return tasks

    @staticmethod
    def get_task(task_id):
        logger.debug('Getting task. %r' % {
            'task_id': task_id,
        })

        task = Task.db.get(Task.column_family, task_id)
        if task:
            if DatabaseObject.validate(Task, task_id, task):
                return Task(id=task_id)

    @staticmethod
    def get_tasks(volume, type=None):
        return Task._get_tasks(volume, type)

    @staticmethod
    def get_failed_tasks(volume, type=None):
        return Task._get_tasks(volume, type, [FAILED])

    @staticmethod
    def get_aborted_tasks(volume, type=None):
        return Task._get_tasks(volume, type, [ABORTED])

    @staticmethod
    def get_pending_tasks(volume, type=None):
        # Valid pending tasks will always be in task_threads if
        # empty assume there are no pending tasks
        if not len(_task_threads):
            return []
        return Task._get_tasks(volume, type, [PENDING, ABORTING])

    @staticmethod
    def get_completed_tasks(volume, type=None):
        return Task._get_tasks(volume, type, [COMPLETE])



from drivnal.constants import *
from drivnal.client import Client
import drivnal.utils as utils
from drivnal import server

@server.app.route('/log/<type>/<volume_id>/<type_id>', methods=['GET'])
def log_get(type, volume_id, type_id):
    client = Client()
    volume = client.get_volume(volume_id)
    data = None

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    if type == 'snapshot':
        snapshot_id = type_id
        snapshot = volume.get_snapshot(int(snapshot_id))

        if not snapshot:
            return utils.jsonify({
                'error': SNAPSHOT_NOT_FOUND,
                'error_msg': SNAPSHOT_NOT_FOUND_MSG,
            }, 404)

        if snapshot.log_size() > MAX_TEXT_SIZE:
            syntax = ''
        else:
            syntax = 'sh'

        data = {
            'id': snapshot.id,
            'syntax': syntax,
            'data': snapshot.log_read(),
        }

    elif type == 'task':
        task_id = type_id
        task = volume.get_task(task_id.encode())

        if not task:
            return utils.jsonify({
                'error': TASK_NOT_FOUND,
                'error_msg': TASK_NOT_FOUND_MSG,
            }, 404)

        if task.log_size() > MAX_TEXT_SIZE:
            syntax = ''
        else:
            syntax = 'sh'

        data = {
            'id': task.id,
            'syntax': syntax,
            'data': task.log_read(),
        }

    if not data:
        return utils.jsonify({
            'error': FILE_NOT_FOUND,
            'error_msg': FILE_NOT_FOUND_MSG,
        }, 404)

    return utils.jsonify(data)



from constants import *
from exec_task import ExecTask
from event import Event
import os
import time
import logging
import subprocess

logger = logging.getLogger(APP_NAME)

class RestoreObject(ExecTask):
    type = RESTORE_OBJECT

    def run(self, objects, destination_path):
        logger.info('Restoring objects. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })

        self.log_path = os.path.join(self.volume.path,
            LOG_DIR, 'restore_%s.log' % int(time.time()))

        for obj in objects:
            args = ['rsync', '--archive', '--hard-links', '--acls',
                '--quiet', '--xattrs',  '--progress', '--super',
                '--log-file-format=%o \"%f\" %l',
                '--log-file=%s' % self.log_path,
                obj.path, destination_path]

            return_code = self._exec(args)
            if return_code is None:
                return

            if return_code != 0:
                raise OSError('Object restore failed, command ' + \
                    'returned non-zero exit status. %r' % {
                        'volume_id': self.volume_id,
                        'snapshot_id': snapshot_id,
                    })

        Event(volume_id=self.volume_id, type=ORIGIN_UPDATED)



from constants import *
from exceptions import *
from exec_task import ExecTask
from event import Event
import os
import time
import shlex
import logging
import subprocess

logger = logging.getLogger(APP_NAME)

class CreateSnapshot(ExecTask):
    type = CREATE_SNAPSHOT

    def remove_snapshot(self):
        self.volume.remove_snapshot(self.snapshot, block=True)

    def _prune_snapshot(self, snapshot):
        keep_log = False
        if snapshot.state == FAILED:
            keep_log = True
        self.volume.remove_snapshot(snapshot, keep_log, True)

    def prune_snapshots(self):
        snapshots = self.volume.get_snapshots()
        failed_snapshots = self.volume.get_failed_snapshots()
        snapshot_count = self.volume.get_snapshot_count()
        snapshot_limit = self.volume.snapshot_limit
        min_free_space = self.volume.min_free_space
        orig_free_space = self.volume.get_space_free()
        prune_count = 0
        max_prune = self.volume.max_prune or DEFAULT_MAX_PRUNE
        max_prune_count = (int(self.orig_snapshot_count * max_prune) -
            (self.orig_snapshot_count - snapshot_count))

        logger.debug('Pruning snapshots. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot.id,
        })

        for snapshot in failed_snapshots:
            logger.info('Auto removing failed snapshot. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot.id,
                'removing_snapshot_id': snapshot.id,
            })
            self._prune_snapshot(snapshot)

        if snapshot_limit:
            for i in xrange((snapshot_count + 1) - snapshot_limit):
                snapshot = snapshots.pop(0)
                logger.info('Exceeding snapshot limit, auto removing ' + \
                    'snapshot. %r' % {
                        'volume_id': self.volume_id,
                        'snapshot_id': self.snapshot.id,
                        'removing_snapshot_id': snapshot.id,
                    })
                self._prune_snapshot(snapshot)

        if not min_free_space:
            return True

        # Remove up to the max prune count of volumes until required free
        # space has been reached
        for snapshot in snapshots:
            if self.volume.get_space_free() >= min_free_space:
                return True
            logger.info('Volume low on space, auto removing snapshot. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot.id,
                'removing_snapshot_id': snapshot.id,
            })
            self._prune_snapshot(snapshot)
            prune_count += 1
            if prune_count >= max_prune_count:
                break

        if self.volume.get_space_free() >= min_free_space:
            return True
        return False

    def _pre_aborted(self):
        self.remove_snapshot()

    def _post_aborted(self):
        Event(type=VOLUMES_UPDATED)

    def run(self):
        self.snapshot = self.volume.SnapshotClass(self.volume)
        self.snapshot_id = self.snapshot.id
        self.orig_snapshot_count = self.volume.get_snapshot_count()
        bandwidth_limit = self.volume.bandwidth_limit
        excludes = self.volume.excludes or []
        source_path = self.volume.source_path
        destination_path = self.snapshot.path
        self.log_path = self.snapshot.log_path
        last_snapshot = self.volume.get_last_snapshot()
        last_snapshot_id = None
        if last_snapshot:
            last_snapshot_id = last_snapshot.id

        logger.info('Creating snapshot. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot.id,
            'last_snapshot_id': last_snapshot_id,
            'task_id': self.id,
        })

        Event(type=VOLUMES_UPDATED)

        args = ['rsync', '--archive', '--delete', '--hard-links',
            '--acls', '--quiet', '--xattrs',  '--progress', '--super',
            '--log-file-format=%o \"%f\" %l', '--log-file=%s' % self.log_path]

        if bandwidth_limit:
            args.append('--bwlimit=%s' % bandwidth_limit)

        rsync_source_path = source_path
        if rsync_source_path[-1] != os.sep:
            rsync_source_path += os.sep

        if rsync_source_path == os.sep:
            logger.debug('Auto excluding default paths for root backup. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot.id,
                'task_id': self.id,
            })
            for exclude in DEFAULT_ROOT_EXCLUDES:
                if exclude in excludes:
                    continue
                excludes.append(exclude)

        auto_excludes = self.volume._get_auto_excludes()
        if auto_excludes:
            excludes += auto_excludes

        for exclude in excludes:
            args.append('--exclude=%s' % exclude)

        setup_args = self.snapshot._setup_snapshot(last_snapshot)
        if setup_args:
            args += setup_args

        args.append(rsync_source_path)
        args.append(destination_path)

        self.prune_snapshots()

        return_code = self._exec(args)
        if return_code is None:
            return

        if return_code != 0 and return_code not in RSYNC_WARN_EXIT_CODES:
            self.snapshot.set_state(FAILED)
            logger.error('Snapshot failed, command ' + \
                'returned non-zero exit status. %r' % {
                    'volume_id': self.volume_id,
                    'snapshot_id': self.snapshot.id,
                    'task_id': self.id,
                    'cmd_args': args,
                    'return_code': return_code,
                })
            raise SnapshotError

        try:
            if return_code == 0:
                self.snapshot.set_state(COMPLETE)
            else:
                self.snapshot.set_state(WARNING)
        except OSError:
            logger.exception('Snapshot failed, unable to rename ' + \
                'snapshot temp directory. %r' % {
                    'volume_id': self.volume_id,
                    'snapshot_id': self.snapshot.id,
                    'task_id': self.id,
                })
            raise SnapshotError

        self.state = COMPLETE
        Event(type=VOLUMES_UPDATED)
        Event(volume_id=self.volume_id, type=SNAPSHOTS_UPDATED)



import optparse
import sys
import os
import drivnal

def drivnal_daemon():
    parser = optparse.OptionParser()
    parser.add_option('-d', '--daemon', action='store_true',
        help='Daemonize process')
    parser.add_option('-p', '--pidfile', type='string',
        help='Path to create pid file')
    parser.add_option('-c', '--conf', type='string',
        help='Path to configuration file')
    (options, args) = parser.parse_args()

    if options.daemon:
        pid = os.fork()
        if pid > 0:
            if options.pidfile:
                with open(options.pidfile, 'w') as pid_file:
                    pid_file.write('%s' % pid)
            sys.exit(0)

    if options.conf:
        conf_path = options.conf
    else:
        conf_path = '/etc/drivnal.conf'

    drivnal.server.conf_path = conf_path
    drivnal.server.run_all()



"""Generate demo data for webapp demo."""
import os
import pprint
import urllib
import mimetypes

objects = []
mime = mimetypes.MimeTypes()

def add_dir(cur_objects, dir_path):
    for dir_name in dir_path.split(os.sep):
        if not dir_name:
            continue

        object_exists = False
        for obj in cur_objects:
            if obj['id'] == dir_name:
                object_exists = True
                cur_objects = obj['objects']
                break

        if not object_exists:
            obj = {
                'id': dir_name,
                'size': None,
                'time': None,
                'type': 'inode/directory',
                'objects': [],
            }
            cur_objects.append(obj)
            cur_objects = obj['objects']

    return cur_objects

def add_path(path):
    dir_paths = []
    dir_paths_dir_names = {}
    dir_paths_file_names = {}

    for (dir_path, dir_names, file_names) in os.walk(path):
        dir_paths.append(dir_path)
        dir_paths_dir_names[dir_path] = dir_names
        dir_paths_file_names[dir_path] = file_names

    for dir_path in sorted(dir_paths):
        cur_objects = add_dir(objects, dir_path)
        dir_names = dir_paths_dir_names[dir_path]
        file_names = dir_paths_file_names[dir_path]

        for dir_name in sorted(dir_names):
            path = os.path.join(dir_path, dir_name)

            object_exists = False
            for obj in cur_objects:
                if obj['id'] == dir_name:
                    object_exists = True
                    break

            if not object_exists:
                obj = {
                    'id': dir_name,
                    'size': None,
                    'time': None,
                    'type': 'inode/directory',
                    'objects': [],
                }
                cur_objects.append(obj)

        for file_name in sorted(file_names):
            path = os.path.join(dir_path, file_name)

            object_exists = False
            for obj in cur_objects:
                if obj['id'] == file_name:
                    object_exists = True
                    break

            if not object_exists:
                try:
                    stat = os.stat(path)
                    size = stat.st_size
                    time = int(stat.st_mtime)
                except OSError:
                    size = None
                    time = None
                obj = {
                    'id': file_name,
                    'size': size,
                    'time': time,
                    'type': mime.guess_type(urllib.pathname2url(path))[0]
                }
                cur_objects.append(obj)

add_path('/home/ubuntu')

#pprint.pprint(objects, indent=0)
print ('%s' % objects).replace(': None', ': null')



from drivnal.constants import *
from drivnal.client import Client
from drivnal.event import Event
import drivnal.utils as utils
from drivnal import server
import time
import uuid

@server.app.route('/event', methods=['GET'])
@server.app.route('/event/<int:last_event>', methods=['GET'])
def event_get(last_event=None):
    if not last_event:
        events = [
            {
                'id': uuid.uuid4().hex,
                'type': 'time',
                'time': int(time.time() * 1000),
            },
        ]
        return utils.jsonify(events)

    run_time = 0
    while run_time <= 30 and not server.interrupt:
        events = []

        for event in Event.get_events(last_event):
            event_data = {
                'id': event.id,
                'type': event.type,
                'time': event.time,
            }
            if event.volume_id:
                event_data['volume_id'] = event.volume_id
            events.append(event_data)

        if len(events):
            return utils.jsonify(events)

        run_time += 0.1
        time.sleep(0.1)

    return utils.jsonify([])



from constants import *
from bucket import Bucket
import utils
import os
import time
import shlex
import datetime
import logging

logger = logging.getLogger(APP_NAME)

class CoreSnapshot(Bucket):
    def __init__(self, volume, dir_name=None):
        self.volume = volume

        if dir_name:
            if dir_name[-7:] == '.%s' % FAILED:
                self.id = int(dir_name[:-7])
                self.state = FAILED
            elif dir_name[-9:] == '.%s' % REMOVING:
                self.id = int(dir_name[:-9])
                self.state = REMOVING
            elif dir_name[-8:] == '.%s' % PENDING:
                self.id = int(dir_name[:-8])
                self.state = PENDING
            elif dir_name[-8:] == '.%s' % WARNING:
                self.id = int(dir_name[:-8])
                self.state = WARNING
            else:
                self.id = int(dir_name)
                self.state = COMPLETE
        else:
            self.id = int(time.time())
            self.state = PENDING

    def __getattr__(self, name):
        if name in ['runtime', 'sent', 'received', 'speed']:
            self.parse_log_file()
        elif name == 'path':
            return self._get_path()
        elif name == 'log_path':
            return self._get_log_path()

        if name not in self.__dict__:
            raise AttributeError(
                'Snapshot instance has no attribute %r' % name)
        return self.__dict__[name]

    def log_read(self):
        try:
            with open(self.log_path) as file_data:
                return file_data.read()
        except OSError, error:
            logger.warning('Failed to read snapshot log file. %r' % {
                'volume_id': self.volume.id,
                'snapshot_id': self.id,
                'error': error,
            })
            return

    def log_size(self):
        try:
            return os.path.getsize(self.log_path)
        except OSError, error:
            logger.warning('Failed to get snapshot log file size. %r' % {
                'volume_id': self.volume.id,
                'snapshot_id': self.id,
                'error': error,
            })
            return

    def parse_log_file(self):
        self.runtime = None
        self.sent = None
        self.received = None
        self.speed = None

        logger.debug('Parsing log file for snapshot. %r' % {
            'volume_id': self.volume.id,
            'snapshot_id': self.id,
        })

        try:
            with open(self.log_path) as log:
                start_time = None
                line = log.readline()

                try:
                    line_split = shlex.split(line)

                    if len(line_split) >= 2:
                        # Get epoch time
                        epoch = line_split[0] + 'T' + line_split[1]
                        epoch = datetime.datetime.strptime(epoch,
                            '%Y/%m/%dT%H:%M:%S')
                        start_time = int(time.mktime(epoch.timetuple()))
                    else:
                        logger.warning('Failed to get snapshot start ' + \
                            'time from log, line split length invalid. %r' % {
                                'volume_id': self.volume.id,
                                'snapshot_id': self.id,
                                'log_line': line,
                            })

                except ValueError:
                    logger.warning('Failed to get snapshot start ' + \
                        'time from log, value error. %r' % {
                            'volume_id': self.volume.id,
                            'snapshot_id': self.id,
                            'log_line': line,
                        })

                # Get last kilobyte of file
                log.seek(0, os.SEEK_END)
                file_size = log.tell()
                log.seek(max(file_size - 1024, 0))
                lines = log.readlines()

                # Find rsync sent command line output
                for line in lines:
                    try:
                        line_split = shlex.split(line)
                    except ValueError:
                        continue

                    if len(line_split) < 10:
                        continue

                    # Get rsync command
                    command = line_split[3]
                    if command == 'sent':
                        if start_time:
                            # Get runtime
                            epoch = line_split[0] + 'T' + line_split[1]
                            epoch = datetime.datetime.strptime(epoch,
                                '%Y/%m/%dT%H:%M:%S')
                            epoch = int(time.mktime(epoch.timetuple()))
                            self.runtime = utils.format_seconds(
                                epoch - start_time)

                        # Get snapshot info
                        try:
                            self.sent = utils.format_bytes(line_split[4])
                        except ValueError:
                            logger.warning('Failed to get sent bytes ' + \
                                'from snapshot log, value error. %r' % {
                                    'volume_id': self.volume.id,
                                    'snapshot_id': self.id,
                                    'log_line': line,
                                })

                        try:
                            self.received = utils.format_bytes(line_split[7])
                        except ValueError:
                            logger.warning('Failed to get received bytes ' + \
                                'from snapshot log, value error. %r' % {
                                    'volume_id': self.volume.id,
                                    'snapshot_id': self.id,
                                    'log_line': line,
                                })

                        try:
                            self.speed = utils.format_bytes(
                                line_split[9]) + '/sec'
                        except ValueError:
                            logger.warning('Failed to get transfer speed ' + \
                                'from snapshot log, value error. %r' % {
                                    'volume_id': self.volume.id,
                                    'snapshot_id': self.id,
                                    'log_line': line,
                                })
        except IOError:
            logger.debug('Failed to read log file for ' + \
                'snapshot, IOError. %r' % {
                    'volume_id': self.volume.id,
                    'snapshot_id': self.id,
                })

    def _get_path(self):
        return ''

    def _get_log_path(self):
        return ''

    def _setup_snapshot(self, last_snapshot):
        pass

    def set_state(self, state):
        if self.state == state:
            return
        self.state = state



import drivnal.utils as utils
from drivnal import server

@server.app.route('/debug/clear_events', methods=['GET'])
def debug_clear_events_get():
    server.app_db.remove('events')
    return 'events_cleared'

@server.app.route('/debug/clear_tasks', methods=['GET'])
def debug_clear_tasks_get():
    server.app_db.remove('tasks')
    return 'tasks_cleared'

@server.app.route('/debug/clear_volumes', methods=['GET'])
def debug_clear_volumes_get():
    server.app_db.remove('system', 'volumes')
    return 'volumes_cleared'

@server.app.route('/debug/dump_system', methods=['GET'])
def debug_dump_system_get():
    return utils.jsonify(server.app_db.get('system'))

@server.app.route('/debug/dump_events', methods=['GET'])
def debug_dump_events_get():
    return utils.jsonify(server.app_db.get('events'))

@server.app.route('/debug/dump_tasks', methods=['GET'])
def debug_dump_tasks_get():
    return utils.jsonify(server.app_db.get('tasks'))



from constants import *
from exceptions import *
from config import Config
from origin import Origin
from task import Task
from event import Event
from core_snapshot import CoreSnapshot
from create_snapshot import CreateSnapshot
from remove_snapshot import RemoveSnapshot
from restore_object import RestoreObject
from move_volume import MoveVolume
import os
import time
import uuid
import logging

logger = logging.getLogger(APP_NAME)

class CoreVolume(Config):
    bool_options = ['email_ssl']
    int_options = ['snapshot_limit', 'bandwidth_limit']
    float_options = ['min_free_space', 'max_prune']
    path_options = ['excludes', 'source_path']
    str_options = ['id', 'name', 'schedule', 'email', 'email_host',
        'email_user', 'email_pass', 'ssh_path', 'ssh_user', 'ssh_key']
    list_options = ['excludes']
    SnapshotClass = CoreSnapshot

    def __init__(self, client, path, create=False):
        try:
            config_path = os.path.join(path, CONF_FILENAME)
        except AttributeError:
            logger.error('Failed to join volume config path. %r' % {
                'volume_id': self.id,
            })
            raise
        Config.__init__(self, config_path)

        self.client = client
        self.orig_path = path
        self.orig_source_path = self.source_path
        self.path = path

        if not self.id:
            if create:
                self.id = uuid.uuid4().hex
                logger.debug('Setting new volume uuid. %r' % {
                    'volume_id': self.id,
                })
                self.commit()
            else:
                raise IOError('Volume doesnt exists')

    def __getattr__(self, name):
        if name == '_snapshot_names':
            self.load_snapshots()
        elif name == 'log_dir':
            return self._get_log_dir()
        return Config.__getattr__(self, name)

    def get_origin(self):
        return Origin(self)

    def get_snapshot(self, id):
        for name in self._snapshot_names:
            snapshot = self.SnapshotClass(self, name)
            if snapshot.id == id:
                return snapshot

    def get_failed_snapshots(self):
        snapshots = []
        for name in self._snapshot_names:
            snapshot = self.SnapshotClass(self, name)
            if snapshot.state == FAILED:
                snapshots.append(snapshot)
        return snapshots

    def get_snapshots(self):
        snapshots = []
        for name in self._snapshot_names:
            snapshot = self.SnapshotClass(self, name)
            if snapshot.state in [REMOVING, WARNING, COMPLETE]:
                snapshots.append(snapshot)
        return snapshots

    def get_snapshot_count(self):
        count = 0
        for name in self._snapshot_names:
            snapshot = self.SnapshotClass(self, name)
            if snapshot.state == COMPLETE:
                count += 1
        return count

    def get_last_snapshot(self):
        for name in reversed(self._snapshot_names):
            snapshot = self.SnapshotClass(self, name)
            if snapshot.state in [WARNING, COMPLETE]:
                return snapshot

    def create_snapshot(self):
        logger.debug('Creating snapshot. %r' % {
            'volume_id': self.id,
        })

        if self.snapshot_pending():
            raise SnapshotAlreadyRunning('Snapshot failed, ' + \
                'snapshot already running. %r' % {
                    'volume_id': self.id,
                })

        logger.debug('Starting snapshot task. %r' % {
            'volume_id': self.id,
        })

        task = CreateSnapshot(volume=self)
        task.start()

        return task

    def remove_snapshot(self, snapshot, keep_log=False, block=False):
        task = RemoveSnapshot(volume=self, snapshot=snapshot)
        task.start(keep_log)

        if block:
            task.join()

        return task

    def snapshot_pending(self):
        tasks = Task.get_pending_tasks(self, CREATE_SNAPSHOT)
        if tasks:
            return True
        return False

    def get_space_used(self):
        return 1 - self.get_space_free()

    def get_task(self, task_id):
        return Task.get_task(task_id)

    def get_tasks(self):
        return Task.get_tasks(self)

    def restore_object(self, objects, destination_path):
        logger.debug('Starting restore object task. %r' % {
            'volume_id': self.id,
        })

        task = RestoreObject(volume=self)
        task.start(objects, destination_path)

        return task

    def load(self):
        try:
            Config.load(self)
        except IOError:
            # Config doesnt exists defaults will be used
            pass

    def load_snapshots(self):
        logger.debug('Loading snapshots. %r' % {
            'volume_id': self.id,
        })

        snapshot_names = []

        for name in self.list_path(SNAPSHOT_DIR, files=False):
            if not name[:8].isdigit() or len(name) < 6:
                continue
            snapshot_names.append(name)
        self._snapshot_names = sorted(snapshot_names)

    def _move_volume(self):
        logger.debug('Starting move volume task. %r' % {
            'volume_id': self.id,
        })

        task = MoveVolume(volume=self)
        task.start()

        return task

    def commit(self):
        Config.commit(self)
        Event(type=VOLUMES_UPDATED)

        if self.orig_path != self.path:
            self._move_volume()

    def _get_log_dir(self):
        return './'

    def _get_auto_excludes(self):
        pass

    def _move_available(self, source_path, destination_path):
        return True

    def _copy_available(self, source_path, destination_path):
        return True

    def get_space_free(self):
        return 1

    def list_path(self, path, files=True, dirs=True):
        return []



from drivnal.constants import *
from drivnal.client import Client
import drivnal.utils as utils
from drivnal import server
import os
import flask

@server.app.route('/path', methods=['GET'])
@server.app.route('/path/<path:path>', methods=['GET'])
def path_get(path=None):
    path = '/' + (path or '')
    paths = []

    if path != '/':
        paths.append({
            'name': '..',
            'path': os.path.abspath(os.path.join(path, os.pardir)),
        })

    try:
        path_list = os.listdir(path)
    except OSError, error:
        return utils.jsonify({
            'error': PATH_NOT_FOUND,
            'error_msg': error.strerror,
        }, 404)

    for name in sorted(path_list):
        full_path = os.path.join(path, name)
        if not os.path.isdir(full_path):
            continue

        paths.append({
            'name': name,
            'path': full_path,
        })

    return utils.jsonify(paths)



from constants import *
from exec_task import ExecTask
from event import Event
from drivnal import server
import os
import time
import shlex
import logging
import shutil
import subprocess

logger = logging.getLogger(APP_NAME)

class MoveVolume(ExecTask):
    type = MOVE_VOLUME

    def run(self):
        source_path = self.volume.orig_path
        destination_path = self.volume.path
        move_available = self.volume._move_available(source_path,
            destination_path)
        copy_available = self.volume._copy_available(source_path, destination_path)

        if not move_available and not copy_available:
            raise OSError('Unable to move volume to specified ' + \
                'location. %r' % {
                    'volume_id': self.volume_id,
                    'task_id': self.id,
                    'source_path': source_path,
                    'destination_path': destination_path,
                })

        logger.debug('Updating client config with volume path. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })
        server.app_db.set('system', 'volumes', self.volume.path, None)

        if move_available:
            logger.info('Moving volume. %r' % {
                'volume_id': self.volume_id,
            })

            source_path += os.sep + '*'

            # String required for bash to expand /*
            args = 'mv %s %s' % (source_path, destination_path)

            return_code = self._exec(args, shell=True)
            if return_code is None:
                return

        else:
            logger.info('Copying volume. %r' % {
                'volume_id': self.volume_id,
            })
            log_path = os.path.join(self.volume.log_dir,
                'copy_volume_%s.log' % int(time.time()))

            rsync_source_path = source_path
            if rsync_source_path[-1] != os.sep:
                rsync_source_path += os.sep

            rsync_destination_path = destination_path
            if rsync_destination_path[-1] != os.sep:
                rsync_destination_path += os.sep

            args = ['rsync', '--archive', '--hard-links', '--acls',
                '--quiet', '--xattrs',  '--progress', '--super',
                '--log-file-format=%o \"%f\" %l', '--log-file=%s' % log_path,
                rsync_source_path, rsync_destination_path]

            return_code = self._exec(args)
            if return_code is None:
                return

            logger.debug('Removing previous copy of volume. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
            })

            for name in os.listdir(source_path):
                path = os.path.join(source_path, name)
                try:
                    if os.path.isfile(path):
                        os.remove(path)
                    else:
                        shutil.rmtree(path)
                except Exception:
                    logger.debug('Failed to delete path while removing ' +
                        'previous volume files. %r' % {
                            'volume_id': self.volume_id,
                            'task_id': self.id,
                        })

        if return_code != 0:
            raise OSError('Volume move failed, command returned ' + \
                'non-zero exit status. Updating client config. %r' % {
                    'volume_id': self.volume_id,
                    'task_id': self.id,
                })
            server.app_db.remove('system', 'volumes', self.volume.path)
        else:
            logger.debug('Volume move complete, updating client config. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
            })
            server.app_db.remove('system', 'volumes', self.volume.orig_path)

        Event(type=VOLUMES_UPDATED)



from drivnal.constants import *
from drivnal.client import Client
import drivnal.utils as utils
from drivnal import server
import flask

@server.app.route('/object/<volume_id>/<snapshot_id>', methods=['GET'])
@server.app.route('/object/<volume_id>/<snapshot_id>/<path:path>',
    methods=['GET'])
def object_get(volume_id, snapshot_id, path=None):
    client = Client()
    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    if snapshot_id == 'origin':
        bucket = volume.get_origin()
    else:
        bucket = volume.get_snapshot(int(snapshot_id))

    if not bucket:
        return utils.jsonify({
            'error': SNAPSHOT_NOT_FOUND,
            'error_msg': SNAPSHOT_NOT_FOUND_MSG,
        }, 404)

    objects_dict = {}
    objects_name = []

    try:
        bucket_objects = bucket.get_objects(path)
    except OSError, error:
        return utils.jsonify({
            'error': PATH_NOT_FOUND,
            'error_msg': error.strerror,
        }, 404)

    for object in bucket_objects:
        objects_dict[object.name] = {
            'id': object.name,
            'type': object.type,
            'syntax': object.syntax,
            'size': object.size,
            'time': object.time,
        }
        objects_name.append(object.name)

    objects = []
    for object_name in sorted(objects_name):
        object = objects_dict[object_name]
        if object['type'] == DIR_MIME_TYPE:
            objects.append(object)

    for object_name in sorted(objects_name):
        object = objects_dict[object_name]
        if object['type'] != DIR_MIME_TYPE:
            objects.append(object)

    return utils.jsonify(objects)



import os
import importlib

for module_name in os.listdir(os.path.dirname(__file__)):
    if module_name == '__init__.py' or module_name[-3:] != '.py':
        continue
    __import__(module_name[:-3], locals(), globals())



import os
import logging
import signal
import time
from constants import *
from database import Database
from config import Config

logger = None

class Server(Config):
    bool_options = ['debug', 'log_debug', 'scheduler']
    int_options = ['port']
    path_options = ['log_path', 'db_path', 'www_path']
    str_options = ['bind_addr']

    def __init__(self):
        Config.__init__(self)
        self._scheduler = None
        self.app = None
        self.app_db = None
        self.mem_db = None
        self.interrupt = False

    def _setup_app(self):
        import flask
        self.app = flask.Flask(APP_NAME)

        global logger
        logger = self.app.logger

    def _setup_conf(self):
        self.set_path(self.conf_path)

    def _setup_log(self):
        if self.log_debug:
            self.log_level = logging.DEBUG
        else:
            self.log_level = logging.INFO

        if self.log_path:
            self.log_handler = logging.FileHandler(self.log_path)
        else:
            self.log_handler = logging.StreamHandler()

        global logger
        if not logger:
            logger = logging.getLogger(APP_NAME)

        logger.setLevel(self.log_level)
        self.log_handler.setLevel(self.log_level)

        self.log_handler.setFormatter(logging.Formatter(
            '[%(asctime)s][%(levelname)s][%(module)s][%(lineno)d] ' +
            '%(message)s'))

        logger.addHandler(self.log_handler)

    def _setup_db(self):
        self.mem_db = Database(None)
        self.app_db = Database(self.db_path or DEFAULT_DB_PATH)

    def _close_db(self):
        self.mem_db.close()
        self.app_db.close()

    def _setup_handlers(self):
        import handlers

    def _setup_static_handler(self):
        www_path = self.www_path or DEFAULT_WWW_PATH

        from werkzeug import SharedDataMiddleware

        self.app.wsgi_app = SharedDataMiddleware(self.app.wsgi_app, {
            '/': os.path.normpath(www_path),
        }, cache=False)

        @self.app.route('/', methods=['GET'])
        def index_get():
            with open(os.path.join(www_path, 'index.html'), 'r') as fd:
                return fd.read()

    def _setup_all(self):
        self._setup_app()
        self._setup_conf()
        self._setup_log()
        self._setup_db()
        self._setup_handlers()
        self._setup_static_handler()

    def _run_wsgi(self):
        import cherrypy.wsgiserver

        logger.info('Starting server...')

        server = cherrypy.wsgiserver.CherryPyWSGIServer(
            (self.bind_addr, self.port), self.app)
        try:
            server.start()
        except (KeyboardInterrupt, SystemExit), exc:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            self.interrupt = True
            self._stop_scheduler()
            logger.info('Stopping server...')
            server.stop()
            self._close_db()

    def _run_wsgi_debug(self):
        logger.info('Starting debug server...')

        # App.run server uses werkzeug logger
        werkzeug_logger = logging.getLogger('werkzeug')
        werkzeug_logger.setLevel(self.log_level)
        werkzeug_logger.addHandler(self.log_handler)

        try:
            self.app.run(host=self.bind_addr, port=self.port, threaded=True)
        finally:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            self.interrupt = True
            self._stop_scheduler()
            # Possible data loss here db closing before debug server
            self._close_db()
            logger.info('Stopping server...')

    def _run_server(self):
        if self.debug:
            self._run_wsgi_debug()
        else:
            self._run_wsgi()

    def _start_scheduler(self):
        logger.info('Starting scheduler...')

        if self.scheduler or self.scheduler is None:
            from scheduler import Scheduler
            self._scheduler = Scheduler()
            self._scheduler.start()

    def _stop_scheduler(self):
        logger.info('Stopping scheduler...')

        if self._scheduler:
            self._scheduler.stop()

    def run_scheduler(self):
        self._setup_conf()
        self._setup_log()
        self._setup_db()

        self._start_scheduler()

        try:
            while True:
                time.sleep(1)
        finally:
            self._stop_scheduler()
            self._close_db()

    def run_server(self):
        self._setup_all()
        self._run_server()

    def run_all(self):
        self._setup_all()
        self._start_scheduler()
        self._run_server()



from drivnal.constants import *
from drivnal.client import Client
from drivnal.task import Task
import drivnal.utils as utils
from drivnal import server
import time

@server.app.route('/snapshot/<volume_id>', methods=['GET'])
def snapshot_get(volume_id):
    client = Client()
    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    snapshots = []
    for snapshot in reversed(volume.get_snapshots()):
        snapshots.append({
            'id': snapshot.id,
            'volume': volume.id,
            'state': snapshot.state,
            'sent': snapshot.sent,
            'received': snapshot.received,
            'speed': snapshot.speed,
            'runtime': snapshot.runtime,
            'has_log': True,
        })

    return utils.jsonify(snapshots)

@server.app.route('/snapshot/<volume_id>', methods=['POST'])
def snapshot_post(volume_id):
    client = Client()
    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    snapshot_task = volume.create_snapshot()

    return utils.jsonify({})

@server.app.route('/snapshot/<volume_id>/<int:snapshot_id>',
    methods=['DELETE'])
def snapshot_delete(volume_id, snapshot_id):
    client = Client()
    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    snapshot = volume.get_snapshot(snapshot_id)

    if not snapshot:
        return utils.jsonify({
            'error': SNAPSHOT_NOT_FOUND,
            'error_msg': SNAPSHOT_NOT_FOUND_MSG,
        }, 404)

    volume.remove_snapshot(snapshot)

    return utils.jsonify({})



"""Change version."""
import fileinput
import sys

VERSION = sys.argv[1]

for line in fileinput.input('../drivnal/__init__.py', inplace=True):
    if '__version__ = ' in line:
        line = '__version__ = \'%s\'' % VERSION
    print line.rstrip('\n')

for line in fileinput.input('../PKGBUILD', inplace=True):
    if 'pkgver=' in line:
        line = 'pkgver=%s' % VERSION
    print line.rstrip('\n')

for line in fileinput.input('../debian/changelog', inplace=True):
    if 'drivnal (' in line:
        line = 'drivnal (%s-0ubuntu1) unstable; urgency=low' % VERSION
    print line.rstrip('\n')



from constants import *
from exceptions import *
from client import Client
from event import Event
from task import Task
from messenger import Messenger
import time
import logging
import threading
import traceback

logger = logging.getLogger(APP_NAME)

class Scheduler:
    def __init__(self):
        self.interrupt = None
        self.thread = None

    def _run(self):
        last_check = time.localtime().tm_min

        while not self.interrupt:
            localtime = time.localtime()

            if localtime.tm_min == last_check:
                time.sleep(1)
                continue
            last_check = localtime.tm_min

            try:
                self.check_schedule(localtime)
            except:
                logger.exception('Scheduler failed to check schedule.')

            try:
                # Clean database every 10 min
                if (int(time.mktime(localtime)) / 60) % 10 == 0:
                    self.clean_database()
            except:
                logger.exception('Scheduler failed to clean database.')

    def _create_snapshot(self, volume):
        try:
            volume.create_snapshot()
        except SnapshotAlreadyRunning:
            logger.debug('Snapshot already running, skipping scheduled ' + \
                'snapshot. %r' % {
                    'volume_id': volume.id,
                })
        except:
            logger.exception('Scheduler failed to call create_snapshot. %r' % {
                'volume_id': volume.id,
            })

    def _check_volume_schedule(self, localtime, volume):
        if not volume.schedule:
            return
        schedule = volume.schedule.lower()

        if schedule == 'none':
            return

        try:
            time_num = int(''.join([x for x in schedule if x.isdigit()]))
        except ValueError:
            time_num = None
        time_unit = ''.join([x for x in schedule if x.isalpha()])

        if time_unit in WEEKDAY_UNITS:
            if (localtime.tm_hour == 4 and
                    localtime.tm_min == 0 and
                    localtime.tm_wday == WEEKDAY_UNITS[time_unit]):
                self._create_snapshot(volume)
        else:
            time_unit = time_unit.replace('s', '')

        if time_unit in MIN_UNITS:
            if (int(time.mktime(localtime)) / 60) % time_num == 0:
                self._create_snapshot(volume)
        elif time_unit in HOUR_UNITS:
            if ((int(time.mktime(localtime)) / 60 / 60) % time_num == 0
                    and localtime.tm_min == 0):
                self._create_snapshot(volume)
        elif time_unit in DAY_UNITS:
            if (localtime.tm_hour == 4 and
                    localtime.tm_min == 0 and
                    ((localtime.tm_yday - 1) % time_num) == 0):
                self._create_snapshot(volume)

    def check_schedule(self, localtime):
        client = Client()

        for volume in client.get_volumes():
            try:
                self._check_volume_schedule(localtime, volume)
            except:
                error_msg = 'Scheduler failed to check volume schedule. %r' % {
                    'volume_id': volume.id
                }
                logger.exception(error_msg)
                msg = Messenger(volume)
                msg.send('%s\n\n%s' % (error_msg, traceback.format_exc()))

    def clean_database(self):
        Task.clean_database()
        Event.clean_database()

    def start(self):
        self.interrupt = None
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def stop(self):
        self.interrupt = True
        self.thread.join()



from setuptools import setup
import os
import sys
import copy
import shlex
import shutil
import fileinput
import drivnal

PATCH_DIR = 'build'
INSTALL_UPSTART = True
INSTALL_SYSTEMD = True

prefix = sys.prefix
for arg in copy.copy(sys.argv):
    if arg.startswith('--prefix'):
        prefix = os.path.normpath(shlex.split(arg)[0].split('=')[-1])
    elif arg == '--no-upstart':
        sys.argv.remove('--no-upstart')
        INSTALL_UPSTART = False
    elif arg == '--no-systemd':
        sys.argv.remove('--no-systemd')
        INSTALL_SYSTEMD = False

if not os.path.exists('build'):
    os.mkdir('build')

data_files = [
    ('/etc', ['data/etc/drivnal.conf']),
    ('/var/lib/drivnal', ['data/var/drivnal.db']),
    ('/var/log', ['data/var/log/drivnal.log']),
    ('/usr/share/drivnal/www', ['www/dist/favicon.ico']),
    ('/usr/share/drivnal/www', ['www/dist/index.html']),
    ('/usr/share/drivnal/www', ['www/dist/robots.txt']),
    ('/usr/share/drivnal/www/css', ['www/dist/css/main.css']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/main.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/require.min.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/worker-coffee.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/worker-css.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/worker-javascript.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/worker-json.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/worker-lua.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/worker-php.js']),
    ('/usr/share/drivnal/www/js', ['www/dist/js/worker-xquery.js']),
]

patch_files = []
if INSTALL_UPSTART:
    patch_files.append('%s/drivnal.conf' % PATCH_DIR)
    data_files.append(('/etc/init', ['%s/drivnal.conf' % PATCH_DIR]))
    shutil.copy('data/init/drivnal.conf', '%s/drivnal.conf' % PATCH_DIR)
if INSTALL_SYSTEMD:
    patch_files.append('%s/drivnal.service' % PATCH_DIR)
    data_files.append(('/etc/systemd/system',
        ['%s/drivnal.service' % PATCH_DIR]))
    shutil.copy('data/systemd/drivnal.service',
        '%s/drivnal.service' % PATCH_DIR)

for file_name in patch_files:
    for line in fileinput.input(file_name, inplace=True):
        line = line.replace('%PREFIX%', prefix)
        print line.rstrip('\n')

setup(
    name='drivnal',
    version=drivnal.__version__,
    description='Simple snapshot backups',
    long_description=open('README.rst').read(),
    author='Zachary Huff',
    author_email='zach.huff.386@gmail.com',
    url='http://drivnal.com/',
    download_url='https://github.com/drivnal/drivnal/archive/%s.tar.gz' % (
        drivnal.__version__),
    keywords='backup, snapshot, web interface, rsync',
    packages=['drivnal', 'drivnal.handlers'],
    license=open('LICENSE').read(),
    zip_safe=False,
    install_requires=[
        'flask',
        'cherrypy>=3.2.0',
    ],
    data_files=data_files,
    entry_points={
        'console_scripts': ['drivnal = drivnal.__main__:drivnal_daemon'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: System :: Archiving :: Backup',
    ],
)



from constants import *
from exec_task import ExecTask
from event import Event
import os
import time
import shlex
import logging
import shutil
import subprocess

logger = logging.getLogger(APP_NAME)

class RemoveSnapshot(ExecTask):
    type = REMOVE_SNAPSHOT

    def _pre_aborted(self):
        self.snapshot.set_state(FAILED)

    def run(self, keep_log=False):
        self.snapshot.set_state(REMOVING)
        Event(volume_id=self.volume_id, type=SNAPSHOTS_UPDATED)

        if not keep_log:
            logger.debug('Removing snapshot log file. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_id': self.id,
            })

            if os.path.isfile(self.snapshot.log_path):
                os.remove(self.snapshot.log_path)

        logger.debug('Removing snapshot directory. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot_id,
            'task_id': self.id,
        })

        args = ['rm', '-rf', self.snapshot.path]
        self._exec(args)
        Event(volume_id=self.volume_id, type=SNAPSHOTS_UPDATED)



from drivnal import database
import threading
import unittest
import requests
import json
import time
import os

URL = 'http://localhost:6500'
HEADERS = {
    'Content-type': 'application/json',
    'Accept': 'application/json',
}
TEMP_DATABSE_PATH = 'drivnal_test.db'


class Database(unittest.TestCase):
    def setUp(self):
        if os.path.isfile(TEMP_DATABSE_PATH):
            os.remove(TEMP_DATABSE_PATH)
        self._db = database.Database(TEMP_DATABSE_PATH)

    def tearDown(self):
        self._db.close()
        os.remove(TEMP_DATABSE_PATH)

    def _fill_column_family(self, num):
        for i in xrange(5):
            for x in xrange(5):
                self._db.set('column_family_%s' % num, 'row_%s' % i,
                    'column_%s' % x, 'value_%s' % x)

        for i in xrange(5):
            for x in xrange(5):
                value = self._db.get('column_family_%s' % num,
                    'row_%s' % i, 'column_%s' % x)
                self.assertEqual(value, 'value_%s' % x)

        for i in xrange(5):
            for x in xrange(5):
                self._db.remove('column_family_%s' % num,
                    'row_%s' % i, 'column_%s' % x)

        for i in xrange(5):
            for x in xrange(5):
                value = self._db.get('column_family_%s' % num,
                    'row_%s' % i, 'column_%s' % x)
                self.assertEqual(value, None)

    def test_database(self):
        for i in xrange(2):
            threads = []
            for x in xrange(300):
                thread = threading.Thread(target=self._fill_column_family,
                    args=(x,))
                thread.start()
                threads.append(thread)

            for thread in threads:
                thread.join()


class Volume(unittest.TestCase):
    def test_vagrant_volume(self):
        #######################################################################
        # Get volume
        #######################################################################
        response = requests.get('%s/volume' % URL, headers=HEADERS)
        data = response.json()

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['id'], 'af9d03d0038d11e38d68080027880ca6')
        self.assertEqual(data[0]['name'], 'Root Backup')
        self.assertEqual(data[0]['source_path'], '/')
        self.assertEqual(data[0]['path'], '/media/backup')
        self.assertEqual(data[0]['excludes'], ['vagrant'])
        self.assertEqual(data[0]['schedule'], '1day')
        self.assertEqual(data[0]['min_free_space'], 0.1)
        self.assertEqual(data[0]['bandwidth_limit'], 0)
        self.assertTrue(isinstance(data[0]['percent_used'], float))

    def test_volume_get_post_put_delete(self):
        #######################################################################
        # Create volume
        #######################################################################
        data = {
            'name': 'Unittest Volume',
            'source_path': '/',
            'path': '/tmp',
            'excludes': ['vagrant'],
            'schedule': '3days',
            'min_free_space': 0.2,
            'bandwidth_limit': 0,
        }
        response = requests.post('%s/volume' % URL, headers=HEADERS,
            data=json.dumps(data))
        data = response.json()

        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {})

        #######################################################################
        # Get volume
        #######################################################################
        response = requests.get('%s/volume' % URL, headers=HEADERS)
        data = response.json()

        self.assertEqual(response.status_code, 200)

        volume_id = None
        for volume in data:
            if volume['name'] == 'Unittest Volume':
                volume = volume
                volume_id = volume['id']
                break

        self.assertIsNotNone(volume_id)
        self.assertEqual(volume['name'], 'Unittest Volume')
        self.assertEqual(volume['source_path'], '/')
        self.assertEqual(volume['path'], '/tmp')
        self.assertEqual(volume['excludes'], ['vagrant'])
        self.assertEqual(volume['schedule'], '3days')
        self.assertEqual(volume['min_free_space'], 0.2)
        self.assertEqual(volume['bandwidth_limit'], 0)
        self.assertTrue(isinstance(volume['percent_used'], float))

        #######################################################################
        # Modify volume
        #######################################################################
        data = {
            'name': 'Unittest Volume 2',
            'source_path': '/home',
            'path': '/media',
            'excludes': ['vagrant', 'lost+found'],
            'schedule': '1day',
            'min_free_space': 0.4,
            'bandwidth_limit': 10240,
        }
        response = requests.put('%s/volume/%s' % (URL, volume_id),
            headers=HEADERS, data=json.dumps(data))
        data = response.json()

        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {})

        # Wait for volume move task to finish
        time.sleep(1)

        #######################################################################
        # Get volume
        #######################################################################
        response = requests.get('%s/volume' % URL, headers=HEADERS)
        data = response.json()

        self.assertEqual(response.status_code, 200)

        volume = None
        for volume in data:
            if volume['id'] == volume_id:
                volume = volume
                break

        self.assertIsNotNone(volume)
        self.assertEqual(volume['id'], volume_id)
        self.assertEqual(volume['name'], 'Unittest Volume 2')
        self.assertEqual(volume['source_path'], '/home')
        self.assertEqual(volume['path'], '/media')
        self.assertEqual(volume['excludes'], ['vagrant', 'lost+found'])
        self.assertEqual(volume['schedule'], '1day')
        self.assertEqual(volume['min_free_space'], 0.4)
        self.assertEqual(volume['bandwidth_limit'], 10240)
        self.assertTrue(isinstance(volume['percent_used'], float))

        #######################################################################
        # Delete volume
        #######################################################################
        response = requests.delete('%s/volume/%s' % (URL, volume['id']),
            headers=HEADERS)
        data = response.json()

        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {})


if __name__ == '__main__':
    unittest.main()



from drivnal.constants import *
from drivnal.client import Client
from drivnal.messenger import Messenger
import drivnal.utils as utils
from drivnal import server
import flask
import os

def _get_volume_object(volume):
    return {
        'id': volume.id,
        'name': volume.name or 'Backup Volume',
        'path': volume.path,
        'source_path': volume.source_path,
        'percent_used': volume.get_space_used(),
        'excludes': volume.excludes or [],
        'schedule': volume.schedule or 0,
        'min_free_space': volume.min_free_space or 0,
        'snapshot_limit': volume.snapshot_limit,
        'bandwidth_limit': volume.bandwidth_limit or 0,
        'snapshot_pending': volume.snapshot_pending(),
        'email': volume.email,
        'email_host': volume.email_host,
        'email_user': volume.email_user,
        'email_pass': volume.email_pass,
        'email_ssl': volume.email_ssl,
    }

@server.app.route('/volume', methods=['GET'])
@server.app.route('/volume/<volume_id>', methods=['GET'])
def volume_get(volume_id=None):
    client = Client()

    if volume_id:
        volume = client.get_volume(volume_id)
        return utils.jsonify(_get_volume_object(volume))

    volumes = []
    for volume in client.get_volumes():
        volumes.append(_get_volume_object(volume))

    return utils.jsonify(volumes)

@server.app.route('/volume', methods=['POST'])
@server.app.route('/volume/<volume_id>', methods=['PUT', 'POST'])
def volume_put_post(volume_id=None):
    client = Client()

    name = flask.request.json['name'].encode()
    source_path = flask.request.json['source_path'].encode()
    path = flask.request.json['path'].encode()
    excludes = flask.request.json['excludes']
    schedule = flask.request.json['schedule']
    min_free_space = flask.request.json['min_free_space']
    snapshot_limit = flask.request.json['snapshot_limit']
    bandwidth_limit = flask.request.json['bandwidth_limit']
    email = flask.request.json['email']
    email_host = flask.request.json['email_host']
    email_user = flask.request.json['email_user']
    email_pass = flask.request.json['email_pass']
    email_ssl = flask.request.json['email_ssl']
    email_send_test = flask.request.json['email_send_test']

    if volume_id:
        volume = client.get_volume(volume_id)
    else:
        volume = client.add_volume(os.path.normpath(path))
        if volume.source_path:
            return utils.jsonify({})

    volume.name = name
    volume.source_path = os.path.normpath(source_path)
    volume.path = os.path.normpath(path)
    volume.excludes = excludes
    for i, exclude in enumerate(volume.excludes):
        volume.excludes[i] = os.path.normpath(exclude)
    volume.schedule = schedule
    volume.min_free_space = min_free_space
    volume.snapshot_limit = snapshot_limit
    volume.bandwidth_limit = bandwidth_limit

    volume.email = email
    volume.email_host = email_host
    volume.email_user = email_user
    volume.email_pass = email_pass
    volume.email_ssl = email_ssl

    volume.commit()

    if email_send_test:
        msg = Messenger(volume)
        msg.send('Test notification.')

    return utils.jsonify({})

@server.app.route('/volume/<volume_id>', methods=['DELETE'])
def volume_delete(volume_id):
    client = Client()
    client.remove_volume(volume_id)

    return utils.jsonify({})



from drivnal.constants import *
from drivnal.client import Client
import drivnal.utils as utils
from drivnal import server

@server.app.route('/text/<volume_id>/<snapshot_id>/<path:path>',
    methods=['GET'])
def text_get(volume_id, snapshot_id, path=None):
    client = Client()
    volume = client.get_volume(volume_id)

    if not volume:
        return utils.jsonify({
            'error': VOLUME_NOT_FOUND,
            'error_msg': VOLUME_NOT_FOUND_MSG,
        }, 404)

    if snapshot_id == 'origin':
        bucket = volume.get_origin()
    else:
        bucket = volume.get_snapshot(int(snapshot_id))

    if not bucket:
        return utils.jsonify({
            'error': SNAPSHOT_NOT_FOUND,
            'error_msg': SNAPSHOT_NOT_FOUND_MSG,
        }, 404)

    object = bucket.get_object(path)
    object.get_syntax()

    syntax = object.syntax
    if object.syntax and object.size > MAX_TEXT_SIZE:
        syntax = ''

    object_data = {
        'id': object.name,
        'type': object.type,
        'syntax': syntax,
        'data': object.read(),
    }

    if not object_data:
        return utils.jsonify({
            'error': FILE_NOT_FOUND,
            'error_msg': FILE_NOT_FOUND_MSG,
        }, 404)

    return utils.jsonify(object_data)



from constants import *
from drivnal import server
import logging

logger = logging.getLogger(APP_NAME)
_RESERVED_ATTRIBUTES = ['column_family', 'bool_columns', 'int_columns',
    'float_columns', 'str_columns', 'cached_columns', 'required_columns',
    'all_columns', 'id']

class DatabaseObject:
    db = server.app_db
    column_family = 'column_family'
    bool_columns = []
    int_columns = []
    float_columns = []
    str_columns = []
    cached_columns = []
    required_columns = []

    def __init__(self):
        self.all_columns = self.bool_columns + self.int_columns + \
            self.float_columns + self.str_columns
        self.id = None

    def __setattr__(self, name, value):
        if name not in _RESERVED_ATTRIBUTES and name in self.all_columns:
            if name in self.cached_columns:
                self.__dict__[name] = value

            if name in self.int_columns or name in self.float_columns:
                self.db.set(self.column_family,
                    self.id, name, str(value))
            elif name in self.bool_columns:
                self.db.set(self.column_family,
                    self.id, name, 't' if value else 'f')
            else:
                self.db.set(self.column_family,
                    self.id, name, value)
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if name in _RESERVED_ATTRIBUTES:
            pass
        elif name in self.all_columns and self.id:
            if name in self.cached_columns and name in self.__dict__:
                return self.__dict__[name]

            value = self.db.get(self.column_family, self.id, name)

            if name in self.int_columns:
                if value:
                    value = int(value)
            elif name in self.float_columns:
                if value:
                    value = float(value)
            elif name in self.bool_columns:
                if value == 't':
                    value = True
                elif value == 'f':
                    value = False

            if name in self.cached_columns:
                self.__dict__[name] = value

            return value
        elif name in self.cached_columns or name not in self.__dict__:
            raise AttributeError('Object instance has no attribute %r' % name)

        return self.__dict__[name]

    @staticmethod
    def validate(db_obj, row, columns):
        valid = True

        for column in db_obj.required_columns:
            if column not in columns:
                valid = False
                break

            if column in db_obj.int_columns:
                try:
                    int(columns[column])
                except ValueError:
                    valid = False
                    break

        if valid:
            if 'remove' in columns:
                logger.debug('Prevented removal of partially complete' + \
                    ' %s. %r' % (db_obj.column_family, {
                        'event_id': row,
                    }))
                db_obj.db.remove(db_obj.column_family, row, 'remove')
        else:
            # Remove broken rows
            if 'remove' in columns:
                logger.info(('Removing broken %s from ' + \
                    'database. %r') % (db_obj.column_family, {
                        '%s_id' % db_obj.column_family: row,
                    }))
                db_obj.db.remove(db_obj.column_family, row)
            else:
                logger.debug(('Queueing removal of broken %s from' + \
                    ' database. %r') % (db_obj.column_family, {
                        '%s_id' % db_obj.column_family: row,
                    }))
                # Its possible row is currently being created and will
                # be valid once created. Wait for next db clean to
                # revalidate and remove row.
                db_obj.db.set(db_obj.column_family, row, 'remove', 'true')

        return valid



from constants import *
import threading
import logging
import copy
import urlparse
import time

logger = logging.getLogger(APP_NAME)

class RedisBackend:
    def __init__(self, host, port, db):
        if host is None:
            host = 'localhost'
        if port is None:
            port = 6379
        if db is None:
            db = 0

        logger.info('Connecting to redis database... %r' % {
            'host': host,
            'port': port,
            'db': db,
        })

        import redis
        self._client = redis.Redis(host=host, port=port, db=db)

    def close(self):
        pass

    def get(self, key):
        value = self._client.get(key)
        if value == '':
            return None
        return value

    def set(self, key, value):
        if value is None:
            value = ''
        if not self._client.set(key, value):
            raise redis.exceptions.ResponseError(
                'Database set returned false.')

    def remove(self, key):
        self._client.delete(key)

    def keys(self, prefix=None):
        if prefix is not None:
            prefix = '%s*' % prefix
        else:
            prefix = '*'
        keys = self._client.keys(prefix)
        return keys

class BerkeleyBackend:
    def __init__(self, db_path):
        logger.info('Opening berkeley database... %r' % {
            'path': db_path,
        })

        try:
            import bsddb3 as bsddb
        except ImportError:
            import bsddb

        self._client = bsddb.db.DB()
        self._client.open(db_path, bsddb.db.DB_HASH, bsddb.db.DB_CREATE)
        self._sync_lock = threading.Lock()
        self._sync_interrupt = False
        self._sync_thread = threading.Thread(target=self._sync_loop)
        self._sync_thread.start()

    def _sync_loop(self):
        while not self._sync_interrupt:
            self.sync()
            time.sleep(0.5)
        self.sync()

    def close(self):
        logger.info('Closing berkeley database...')
        self._sync_interrupt = True
        self._sync_thread.join()

    def get(self, key):
        return self._client.get(key=key)

    def set(self, key, value):
        self._client.put(key=key, data=value)

    def remove(self, key):
        self._client.delete(key=key)

    def keys(self, prefix):
        return self._client.keys()

    def sync(self):
        if self._sync_lock.locked():
            return
        self._sync_lock.acquire()
        try:
            self._client.sync()
        finally:
            self._sync_lock.release()

class MemoryBackend:
    def __init__(self):
        logger.info('Creating memory database...')
        self._data = {}

    def close(self):
        pass

    def get(self, key):
        if key in self._data:
            return self._data[key]

    def set(self, key, value):
        self._data[key] = value

    def remove(self, key):
        if key in self._data:
            del self._data[key]

    def keys(self, prefix):
        keys = []
        for key in self._data:
            keys.append(key)
        return keys

class Database:
    def __init__(self, db_path):
        self._db_lock = threading.Lock()

        if db_path is None or db_path == 'none':
            self._db = MemoryBackend()
        elif db_path.startswith('redis://'):
            parse = urlparse.urlparse(db_path)
            if len(parse.netloc.split(':')) > 1:
                host = parse.netloc.split(':')[0]
                port = int(parse.netloc.split(':')[1])
            else:
                if parse.netloc:
                    host = parse.netloc
                else:
                    host = None
                port = None
            if parse.path:
                db = int(parse.path.replace('/', ''))
            else:
                db = None
            self._db = RedisBackend(host, port, db)
        else:
            self._db = BerkeleyBackend(db_path)

    def _get(self, key):
        self._db_lock.acquire()
        try:
            value = self._db.get(key)
        finally:
            self._db_lock.release()
        return value

    def _prefix_get(self, prefix):
        prefix_len = len(prefix)
        keys = {}

        self._db_lock.acquire()
        try:
            for key in self._db.keys(prefix):
                if key.startswith(prefix):
                    keys[key[prefix_len:]] = self._db.get(key)
        finally:
            self._db_lock.release()

        return keys

    def _set(self, key, value):
        self._db_lock.acquire()
        try:
            self._db.set(key, value)
        finally:
            self._db_lock.release()

    def _remove(self, key):
        self._db_lock.acquire()
        try:
            self._db.remove(key)
        except bsddb.db.DBNotFoundError:
            pass
        finally:
            self._db_lock.release()

    def _prefix_remove(self, prefix):
        prefix_len = len(prefix)

        self._db_lock.acquire()
        try:
            for key in self._db.keys(prefix):
                if key.startswith(prefix):
                    self._db.remove(key)
        finally:
            self._db_lock.release()

    def close(self):
        self._db.close()

    def get(self, column_family, row=None, column=None):
        if not row and column:
            raise TypeError('Must specify a row for column')

        if not row:
            keys = {}
            prefix_keys = self._prefix_get('%s-' % column_family)

            for key in prefix_keys:
                key_split = key.split('-', 1)
                if key_split[0] not in keys:
                    keys[key_split[0]] = {}
                keys[key_split[0]][key_split[1]] = prefix_keys[key]

            return keys

        if not column:
            return self._prefix_get('%s-%s-' % (column_family, row))

        key_name = '%s-%s-%s' % (column_family, row, column)
        return self._get(key_name)

    def set(self, column_family, row, column, value):
        self._set('%s-%s-%s' % (column_family, row, column), value)

    def remove(self, column_family, row=None, column=None):
        if not row and column:
            raise TypeError('Must specify a row for column')

        key_name = '%s-' % column_family
        if row:
            key_name += '%s-' % row

        if not column:
            self._prefix_remove(key_name)
        else:
            key_name += column
            self._remove(key_name)



from constants import *
from core_volume import CoreVolume
from remote_snapshot import RemoteSnapshot
import os
import logging

logger = logging.getLogger(APP_NAME)

class RemoteVolume(CoreVolume):
    SnapshotClass = RemoteSnapshot

    def _get_log_dir(self):
        return './'

    def _get_auto_excludes(self):
        pass

    def _move_available(self, source_path, destination_path):
        return True

    def _copy_available(self, source_path, destination_path):
        return True

    def get_space_free(self):
        return 1

    def list_path(self, path, files=True, dirs=True):
        return []



from constants import *
from core_snapshot import CoreSnapshot
import logging

logger = logging.getLogger(APP_NAME)

class RemoteSnapshot(CoreSnapshot):
    def _get_path(self):
        dir_name = str(self.id)
        if self.state != COMPLETE:
            dir_name = '%s.%s' % (dir_name, self.state)
        return '%s@%s%s' % (self.volume.ssh_user, self.volume.ssh_path,
            os.sep + os.path.join(SNAPSHOT_DIR, dir_name) + os.sep)

    def _get_log_path(self):
        return os.path.join(self.volume.log_dir, 'snapshot_%s.log' % self.id)

    def _setup_snapshot(self, last_snapshot):
        pass

    def set_state(self, state):
        if self.state == state:
            return
        self.state = state
